/* Core library for the Barons strategic engine */

#include <stdio.h>
#include "core.h"
#include "file.h"

/* File I/O for Things */

static thing_data *read_thing(FILE *f, char *firstline, int* total_read) {
  char *curr;
  thing_data *thing = 0;
  thing_data *subthing = 0;
  thing_type type;
  int things_read;
  char *tag, *value;
  char dataline[256];
  char reading[256] = "";

  *total_read = 0;

  /* Find out what sort of Thing we've got */
  for (type = tt_max ; type >= tt_min ; type--) {
    if (thing_abbr[type] == firstline[0]) break;
  }
  if (type == tt_unknown) return 0;

  /* Allocate storage for the Thing */
  thing = (thing_data*)calloc(1, sizeof(thing_data));
  if (!thing) return 0;   /* In case allocation fails */
  thing->type = type;
  *total_read = 1;

  /* Read lines and parse them according to the type of Thing being read */
  while (fgets(dataline, sizeof(dataline), f)) {
    strtrim(dataline);
    if (dataline[0] == '}') break;      /* Check for end of Thing */
    if (dataline[0] == '#') continue;   /* Skip comments */

    /* Attempt to separate line into tag and value parts */
    tag = value = dataline;
    while ((*value) && (*value != ' ')) value++;    /* Find first space */
    /* If we found a space, change it to a null and advance to the next
       character.  tag is now the first word of the line and value is the
       rest */
    if (*value) *value++ = '\0';

    /* If we're reading a list, find out whether we've got the start of a new
       Thing */
    if (reading[0]) {
      /* Check for { */
      if (value[0] == '{') 
        tag = reading;        /* So the appropriate code will catch it */
      else
        reading[0] = '\0';    /* List has ended */
    }

    /* Test whether this is a name attribute, since that's there regardless
       of Thing type */
    if (!strcmp(tag, "name")) {
      strcpy(thing->data.generic.name, value);
      continue;
    }

    /* Now for the fun part...  Listing all the recognized values for each
       type and specifying how to read them */
    switch (type) {
      case tt_resource_template : {
        if (!strcmp(tag, "influence")) {
          if (!reading[0]) strcpy(reading, tag);
          else if (subthing = read_thing(f, dataline, &things_read)) {
            add_to_list(&thing->data.resource_template.influence, subthing);
            *total_read += things_read;
          }
        } else if (!strcmp(tag, "can_overspend")) {
          thing->data.resource_template.can_overspend = atoi(value);
        } else if (!strcmp(tag, "decay_rate")) {
          thing->data.resource_template.decay_rate = atoi(value);
        } else if (!strcmp(tag, "decay_when")) {
          thing->data.resource_template.decay_when =
                            (strcmp(value, "early") ? dt_late : dt_early);
        } else if (!strcmp(tag, "decay_to")) {
          if (!reading[0]) strcpy(reading, tag);
          else if (subthing = read_thing(f, dataline, &things_read)) {
            add_to_list(&thing->data.resource_template.decay_to, subthing);
            *total_read += things_read;
          }
        }

        break;
      }
      
      case tt_resource : {
        if (!strcmp(tag, "type")) {
          thing->data.resource.type =
                            &find_any_thing(value)->data.resource_template;
          if (!thing->data.resource.name[0])
            strcpy(thing->data.resource.name, value);
        } else if (!strcmp(tag, "qty")) {
          thing->data.resource.qty = atoi(value);
        }

        break;
      }
      
      case tt_race : {
        if (!strcmp(tag, "food_consumption")) {
          thing->data.race.food_consumption = atoi(value);
        } else if (!strcmp(tag, "mana_consumption")) {
          thing->data.race.mana_consumption = atoi(value);
        } else if (!strcmp(tag, "resource_consumption")) {
          if (!reading[0]) strcpy(reading, tag);
          else if (subthing = read_thing(f, dataline, &things_read)) {
            add_to_list(&thing->data.race.resource_consumption, subthing);
            *total_read += things_read;
          }
        } else if (!strcmp(tag, "might_production")) {
          thing->data.race.might_production = atoi(value);
        } else if (!strcmp(tag, "labor_production")) {
          thing->data.race.labor_production = atoi(value);
        } else if (!strcmp(tag, "knowledge_production")) {
          thing->data.race.knowledge_production = atoi(value);
        } else if (!strcmp(tag, "resource_production")) {
          if (!reading[0]) strcpy(reading, tag);
          else if (subthing = read_thing(f, dataline, &things_read)) {
            add_to_list(&thing->data.race.resource_production, subthing);
            *total_read += things_read;
          }
        } else if (!strcmp(tag, "growth_rate")) {
          thing->data.race.growth_rate = atoi(value);
        } else if (!strcmp(tag, "wanderlust")) {
          thing->data.race.wanderlust = atoi(value);
        }

        break;
      }
      
      case tt_population : {
        if (!strcmp(tag, "race")) {
          thing->data.population.race = &find_any_thing(value)->data.race;
          if (!thing->data.population.name[0])
            strcpy(thing->data.population.name, value);
        } else if (!strcmp(tag, "size")) {
          thing->data.population.size = atoi(value);
        } else if (!strcmp(tag, "skill")) {
          thing->data.population.skill = atoi(value);
        } else if (!strcmp(tag, "happiness")) {
          thing->data.population.happiness = atoi(value);
        }

        break;
      }
      
      case tt_factory_template : {
        if (!strcmp(tag, "input")) {
          if (!reading[0]) strcpy(reading, tag);
          else if (subthing = read_thing(f, dataline, &things_read)) {
            add_to_list(&thing->data.factory_template.input, subthing);
            *total_read += things_read;
          }
        } else if (!strcmp(tag, "output")) {
          if (!reading[0]) strcpy(reading, tag);
          else if (subthing = read_thing(f, dataline, &things_read)) {
            add_to_list(&thing->data.factory_template.output, subthing);
            *total_read += things_read;
          }
        } else if (!strcmp(tag, "capacity")) {
          thing->data.factory_template.capacity = atoi(value);
        } else if (!strcmp(tag, "required")) {
          if (!reading[0]) strcpy(reading, tag);
          else if (subthing = read_thing(f, dataline, &things_read)) {
            add_to_list(&thing->data.factory_template.required, subthing);
            *total_read += things_read;
          }
        } else if (!strcmp(tag, "cost")) {
          if (!reading[0]) strcpy(reading, tag);
          else if (subthing = read_thing(f, dataline, &things_read)) {
            add_to_list(&thing->data.factory_template.cost, subthing);
            *total_read += things_read;
          }
        } else if (!strcmp(tag, "repair_cost")) {
          if (!reading[0]) strcpy(reading, tag);
          else if (subthing = read_thing(f, dataline, &things_read)) {
            add_to_list(&thing->data.factory_template.repair_cost, subthing);
            *total_read += things_read;
          }
        } else if (!strcmp(tag, "depletion")) {
          thing->data.factory_template.depletion = atoi(value);
        }

        break;
      }
      
      case tt_factory : {
        if (!strcmp(tag, "type")) {
          thing->data.factory.type =
                                &find_any_thing(value)->data.factory_template;
          if (!thing->data.factory.name[0])
            strcpy(thing->data.factory.name, value);
        } else if (!strcmp(tag, "qty")) {
          thing->data.factory.qty = atoi(value);
        } else if (!strcmp(tag, "capacity")) {
          thing->data.factory.capacity = atoi(value);
        }

        break;
      }
      
      case tt_site_template : {
        if (!strcmp(tag, "affects")) {
          if (!reading[0]) strcpy(reading, tag);
          else if (subthing = read_thing(f, dataline, &things_read)) {
            add_to_list(&thing->data.site_template.affects, subthing);
            *total_read += things_read;
          }
        } else if (!strcmp(tag, "magnitude")) {
          thing->data.site_template.magnitude = atoi(value);
        }

        break;
      }
      
      case tt_site : {
        if (!strcmp(tag, "type")) {
          thing->data.site.type = &find_any_thing(value)->data.site_template;
          if (!thing->data.site.name[0])
            strcpy(thing->data.site.name, value);
        }

        break;
      }

      case tt_allocation : {
        if (!strcmp(tag, "resource")) {
          if (!reading[0]) strcpy(reading, tag);
          else if (subthing = read_thing(f, dataline, &things_read)) {
            add_to_list(&thing->data.allocation.resource, subthing);
            *total_read += things_read;
          }
        } else if (!strcmp(tag, "type")) {
          if (!strcmp(value, "quantity")) {
            thing->data.allocation.type = al_quantity;
          } else if (!strcmp(value, "percent")) {
            thing->data.allocation.type = al_percent;
          } else if (!strcmp(value, "max")) {
            thing->data.allocation.type = al_max;
          }
        }

        break;
      }

     case tt_construction : {
        if (!strcmp(tag, "factory")) {
          thing->data.construction.factory = 
                          &find_any_thing(value)->data.factory_template;
          if (!thing->data.construction.name[0])
            strcpy(thing->data.construction.name, value);
        } else if (!strcmp(tag, "remaining_cost")) {
          if (!reading[0]) strcpy(reading, tag);
          else if (subthing = read_thing(f, dataline, &things_read)) {
            add_to_list(&thing->data.construction.remaining_cost, subthing);
            *total_read += things_read;
          }
        }

        break;
      }

      case tt_territory : {
        if (!strcmp(tag, "populace")) {
          if (!reading[0]) strcpy(reading, tag);
          else if (subthing = read_thing(f, dataline, &things_read)) {
            add_to_list(&thing->data.territory.populace, subthing);
            *total_read += things_read;
          }
        } else if (!strcmp(tag, "sites")) {
          if (!reading[0]) strcpy(reading, tag);
          else if (subthing = read_thing(f, dataline, &things_read)) {
            add_to_list(&thing->data.territory.sites, subthing);
            *total_read += things_read;
          }
        } else if (!strcmp(tag, "factories")) {
          if (!reading[0]) strcpy(reading, tag);
          else if (subthing = read_thing(f, dataline, &things_read)) {
            add_to_list(&thing->data.territory.factories, subthing);
            *total_read += things_read;
          }
        } else if (!strcmp(tag, "resources")) {
          if (!reading[0]) strcpy(reading, tag);
          else if (subthing = read_thing(f, dataline, &things_read)) {
            add_to_list(&thing->data.territory.resources, subthing);
            *total_read += things_read;
          }
        } else if (!strcmp(tag, "allocated")) {
          if (!reading[0]) strcpy(reading, tag);
          else if (subthing = read_thing(f, dataline, &things_read)) {
            add_to_list(&thing->data.territory.allocated, subthing);
            *total_read += things_read;
          }
        } else if (!strcmp(tag, "building")) {
          if (!reading[0]) strcpy(reading, tag);
          else if (subthing = read_thing(f, dataline, &things_read)) {
            add_to_list(&thing->data.territory.building, subthing);
            *total_read += things_read;
          }
        }

        break;
      }
    }
  }

  return thing;
}

int read_everything(FILE *f) {
  thing_data *new_thing;
  int result = 0;
  int things_read;
  char dataline[256];
  char *curr;
  
  while (fgets(dataline, sizeof(dataline), f)) {
    /* Get a line from the file and try to read a Thing starting on it */
    strtrim(dataline);
    if (dataline[0] == '#') continue;   /* Skip lines starting with # */

    /* Check for { */
    curr = dataline;
    while ((*curr) && (*curr != '{')) curr++;
    if (!*curr) continue;

    if (new_thing = read_thing(f, dataline, &things_read)) {
      add_to_master(&new_thing);
      result += things_read;
    }
  }

  return result;
}

static int indent = 0;   /* How far write_thing should indent */

static int write_list(FILE *f, const thing_data *list, char* title) {
  char pad[16] = "               ";         /* Spaces for indentation */
  const thing_data *curr = list;
  int result = 0;

  if (!list) return 0;

  pad[indent] = '\0';       /* Set to proper length */
  fprintf(f, "%s%s\n", pad, title);

  while (curr) {
    result += write_thing(f, curr);
    curr = next_thing(curr);
  }

  return result;
}

int write_thing(FILE *f, const thing_data *thing) {
  int result = 1;
  char pad[16] = "               ";         /* Spaces for indentation */

  /* Skip resources that aren't there any more */
  if ((thing->type == tt_resource) && (thing->data.resource.qty == 0))
    return 0;

  pad[indent] = '\0';       /* Set to proper length */

  /* Opening data, independent of Thing type */
  fprintf(f, "%s%c {\n", pad, thing_abbr[thing->type]);
  pad[indent] = ' '; pad[indent + 2] = '\0';
  indent += 2;

  /* Type-dependent data */
  switch (thing->type) {
    case tt_unknown : {
      fprintf(f, "%sname %s\n", pad, thing->data.generic.name);

      break;
    }

    case tt_resource_template : {
      fprintf(f, "%sname %s\n", pad, thing->data.resource_template.name);
      result += write_list(f, thing->data.resource_template.influence,
                              "influence");
      
      fprintf(f, "%scan_overspend %d\n", pad, 
                 thing->data.resource_template.can_overspend);

      fprintf(f, "%sdecay_rate %d\n", pad, 
                 thing->data.resource_template.decay_rate);
      if (thing->data.resource_template.decay_when != dt_late) {
        fprintf(f, "%sdecay_when ", pad);
        switch(thing->data.resource_template.decay_when) {
          case dt_early : { fprintf(f, "early\n");    break; }
          default:        { fprintf(f, "late\n");     break; }
        }
      }
      result += write_list(f, thing->data.resource_template.decay_to,
                              "decay_to");
      
      break;
    }

    case tt_resource : {
      if (strcmp(thing->data.resource.name, thing->data.resource.type->name))
        fprintf(f, "%sname %s\n", pad, thing->data.resource.name);
      fprintf(f, "%stype %s\n", pad, thing->data.resource.type->name);
      fprintf(f, "%sqty %d\n", pad, thing->data.resource.qty);

      break;
    }

    case tt_race : {
      fprintf(f, "%sname %s\n", pad, thing->data.race.name);
      fprintf(f, "%sfood_consumption %d\n", pad,
                 thing->data.race.food_consumption);
      fprintf(f, "%smana_consumption %d\n", pad,
                 thing->data.race.mana_consumption);
      result += write_list(f, thing->data.race.resource_consumption,
                              "resource_consumption");
      
      fprintf(f, "%smight_production %d\n", pad,
                 thing->data.race.might_production);
      fprintf(f, "%slabor_production %d\n", pad,
                 thing->data.race.labor_production);
      fprintf(f, "%sknowledge_production %d\n", pad,
                 thing->data.race.knowledge_production);
      result += write_list(f, thing->data.race.resource_production,
                              "resource_production");

      fprintf(f, "%sgrowth_rate %d\n", pad, thing->data.race.growth_rate);
      fprintf(f, "%swanderlust %d\n", pad, thing->data.race.wanderlust);

      break;
    }

    case tt_population : {
      if (strcmp(thing->data.population.name,
                                      thing->data.population.race->name))
        fprintf(f, "%sname %s\n", pad, thing->data.population.name);
      fprintf(f, "%srace %s\n", pad, thing->data.population.race->name);

      fprintf(f, "%ssize %d\n", pad, thing->data.population.size);
      fprintf(f, "%sskill %d\n", pad, thing->data.population.skill);
      fprintf(f, "%shappiness %d\n", pad, thing->data.population.happiness);

      break;
    }

    case tt_factory_template : {
      fprintf(f, "%sname %s\n", pad, thing->data.factory_template.name);
      result += write_list(f, thing->data.factory_template.input, "input");
      result += write_list(f, thing->data.factory_template.output, "output");
      fprintf(f, "%scapacity %d\n", pad,
                 thing->data.factory_template.capacity);

      result += write_list(f, thing->data.factory_template.required,
                              "required");
      result += write_list(f, thing->data.factory_template.cost, "cost");
      result += write_list(f, thing->data.factory_template.repair_cost,
                              "repair_cost");

      fprintf(f, "%sdepletion %d\n", pad,
                 thing->data.factory_template.depletion);

      break;
    }

    case tt_factory : {
      if (strcmp(thing->data.factory.name, thing->data.factory.type->name))
        fprintf(f, "%sname %s\n", pad, thing->data.factory.name);
      fprintf(f, "%stype %s\n", pad, thing->data.factory.type->name);

      fprintf(f, "%sqty %d\n", pad, thing->data.factory.qty);
      fprintf(f, "%scapacity %d\n", pad, thing->data.factory.capacity);

      break;
    }

    case tt_site_template : {
      fprintf(f, "%sname %s\n", pad, thing->data.site_template.name);
      result += write_list(f, thing->data.site_template.affects, "affects");
      fprintf(f, "%smagnitude %d\n", pad,
                 thing->data.site_template.magnitude);

      break;
    }

    case tt_site : {
      if (strcmp(thing->data.site.name, thing->data.site.type->name))
        fprintf(f, "%sname %s\n", pad, thing->data.site.name);
      fprintf(f, "%stype %s\n", pad, thing->data.site.type->name);

      break;
    }

    case tt_allocation : {
      fprintf(f, "%sname %s\n", pad, thing->data.allocation.name);
      result += write_list(f, thing->data.allocation.resource, "resource");
      fprintf(f, "%stype ", pad);
      switch (thing->data.allocation.type) {
        case al_quantity  : { printf("quantity\n");   break; }
        case al_percent   : { printf("percent\n");    break; }
        case al_max       : { printf("max\n");        break; }
      }
      break;
    }

    case tt_construction : {
      if (strcmp(thing->data.construction.name,
                 thing->data.construction.factory->name))
        fprintf(f, "%sname %s\n", pad, thing->data.construction.name);
      fprintf(f, "%sfactory %s\n", pad,
                                   thing->data.construction.factory->name);
      write_list(f, thing->data.construction.remaining_cost,
                 "remaining_cost");
      break;
    }

    case tt_territory : {
      fprintf(f, "%sname %s\n", pad, thing->data.territory.name);
      result += write_list(f, thing->data.territory.populace, "populace");
      result += write_list(f, thing->data.territory.sites, "sites");
      result += write_list(f, thing->data.territory.factories, "factories");
      result += write_list(f, thing->data.territory.resources, "resources");
      result += write_list(f, thing->data.territory.allocated, "allocated");
      result += write_list(f, thing->data.territory.building, "building");

      break;
    }
  }

  /* Closing brace */
  indent -= 2;
  pad[indent] = '\0';
  fprintf(f, "%s}\n", pad);

  return result;
}

int write_everything(FILE *f) {
  int result = 0;
  thing_type type;
  char title[32];

  for (type = tt_min ; type <= tt_max ; type++) {
    sprintf(title, "\n# %s", thing_name[type]);
    result += write_list(f, find_thing(type, ""), title);
  }

  return result;
}
